#!/usr/bin/env python2.7
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""iqfeed: Data downloader for Iqfeed/DTN
      Tibor Kiss <tibor.kiss@gmail.com> - Copyright (c) 2012-2016 All rights reserved

Usage:
  iqfeed process-file <filename> <start_year> <end_year> [-d DIR] [-i CON] [-t TZ] [-D]
  iqfeed download <instrument> <start_year> <end_year> [-d DIR] [-i CON] [-t TZ] [-D]
  iqfeed -h | --help

Commands:
  download            Download the specified instrument
  get-from-file       Download instruments listed in the specified file

Options:
  -d DIR --download-dir DIR   Directory where the files will be downloaded [default: .]
  -i CON --iqfeed CON         IQFeed host & port [default: localhost:9100]
  -t TZ --tz TZ               Time zone [default: US/Eastern]
  -D                          Debug mode
  -h                          Help screen

Note:
Date format for end_date and start_date: YYYYMMDD

"""


import os
import sys
import logging

import docopt
import pytz

from .download import get_bars
from .tools import get_instruments_from_file, write_bars_to_file


def main():
    args = docopt.docopt(__doc__, argv=sys.argv[1:], version=0.4)

    log = logging.getLogger()
    log_console = logging.StreamHandler(sys.stdout)
    log.setLevel(logging.DEBUG if args['-D'] else logging.INFO)
    log_console.setLevel(logging.DEBUG if args['-D'] else logging.INFO)
    log.addHandler(log_console)

    if args['download']:
        instruments = (args['<instrument>'], )
    elif args['process-file']:
        instruments = get_instruments_from_file(args['<filename>'])

    start_year = int(args['<start_year>'])
    end_year = int(args['<end_year>'])
    iqfeed_host, iqfeed_port = args['--iqfeed'].split(':')
    iqfeed_port = int(iqfeed_port)
    tz = pytz.timezone(args['--tz'])
    download_dir = args['--download-dir']
    seconds_per_bar = 60  # 1M data

    for (i, instrument) in enumerate(instruments):
        try:
            log.info("Processing %s (%d out of %d)", instrument, i+1, len(instruments))

            for year in range(start_year, end_year+1):
                filename = '%s/HC-%s-1M-%d-iqfeed.csv.gz' % (download_dir, instrument, year)
                start_date = '%s0101' % year
                end_date = '%s1231' % year

                if os.path.exists(filename):
                    log.info('File already exists: %s', filename)
                    continue
                else:
                    log.info('Downloading to %s', filename)

                bars = get_bars(instrument, start_date, end_date, tz, seconds_per_bar, iqfeed_host, iqfeed_port)
                if len(bars):
                    write_bars_to_file(bars, filename, tz)

        except Exception as e:
            log.error('Exception during download, continuing', exc_info=e)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import socket
import contextlib
from collections import namedtuple
from datetime import datetime
import logging

from backports.functools_lru_cache import lru_cache

from .tools import retry

log = logging.getLogger(__name__)

Bar = namedtuple('IQFeedBar', ['datetime', 'open', 'high', 'low', 'close', 'volume'])


def __download_historical_data(iqfeed_socket, chunk_size=65535):
    """
    Read the data from iqfeed_socket with the given chunk size.
    The collected data is returned as a string or exception is raised on error
    """
    buffer_ = ""
    chunk = ""
    end_msg = '\n!ENDMSG!,\r\n'

    while not chunk.endswith(end_msg):
        chunk = iqfeed_socket.recv(chunk_size)

        if chunk.startswith('E,'):  # Error condition
            if chunk.startswith('E,!NO_DATA!'):
                log.warn('No data available for the given instrument')
                break
            else:
                raise Exception(chunk)
        log.debug('New chunk: %s', " ".join("{:02x}".format(ord(c)) for c in chunk[-1*len(end_msg):]))

        buffer_ += chunk

    # Remove the end message string
    buffer_ = buffer_[:-1 * len(end_msg)]

    # Cut off CR
    buffer_ = buffer_.replace('\r', '')

    return buffer_


@lru_cache(maxsize=780000)  # 10 years worth of datetimes
def __create_datetime(datetime_str, format_str, timezone):
    # It takes a reasonable amount of time to construct the datetime fields with timezone info.
    # This function is used to cache the datetime object results.
    dt = datetime.strptime(datetime_str, format_str)
    return timezone.localize(dt)


@retry(5, delay=2)
def get_bars(instrument, start_date, end_date, tz, seconds_per_bar,
             iqfeed_host='localhost', iqfeed_port=9100, timeout=5.0):
    """
    Returns list of Bar instances for the given instrument, time period, time zone and bar frequency (second_per_bar).
    The function is retried 5 times in 5 second intervals to alleviate IQFeed daemon's glitches.
    """
    # IQFeed accepts messages in the following format:
    #   CMD,SYM,[options]\n.
    # Notice the newline character. This must be added otherwise the request will not work.
    # The provided options are
    #   [second_per_bar],[beginning date: CCYYMMDD HHmmSS],[ending date: CCYYMMDD HHmmSS],[empty],
    #   [beginning time filter: HHmmSS],[ending time filter: HHmmSS],[old or new: 0 or 1],[empty],
    #   [second per data point].
    #
    # Source: https://github.com/bwlewis/iqfeed/blob/master/man/HIT.Rd
    begin_time_filter = '092900'
    end_time_filter = '155900'
    historical_data_request = "HIT,%s,%s,%s,%s,,%s,%s,1\n" % (instrument, seconds_per_bar, start_date, end_date,
                                                              begin_time_filter, end_time_filter)

    log.debug("IQFeed historical data request: %s", historical_data_request.rstrip())

    # Open a streaming socket to the IQFeed daemon
    with contextlib.closing(socket.create_connection((iqfeed_host, iqfeed_port))) as iqfeed_socket:
        iqfeed_socket.settimeout(timeout)

        # Send the historical data request historical_data_request and buffer the data
        iqfeed_socket.sendall(historical_data_request)
        data = __download_historical_data(iqfeed_socket)

    bars = []
    if len(data):
        for line in data.split('\n'):
            # Returned fields in data are: datetime, high, low, open, close, volume, XXX?, YYYY?
            (datetime_str, high, low, open_, close, volume, _, _) = line.split(',')
            if volume.find('.') != -1:
                raise Exception("Float as a volume, strange...: %s" % line)

            log.debug("%s open=%s high=%s low=%s close=%s volumes=%s", datetime_str, high, low, open_, close, volume)
            dt = __create_datetime(datetime_str, format_str="%Y-%m-%d %H:%M:%S", timezone=tz)
            (open_, high, low, close, volume) = (float(open_), float(high), float(low), float(close), int(volume))

            bar = Bar(dt, float(open_), float(high), float(low), float(close), int(volume))
            bars.append(bar)

    log.debug("Returning %d bars", len(bars))

    return bars



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import wraps
import logging
import time
import csv
import os


log = logging.getLogger(__name__)


def retry(tries, exceptions=None, delay=0):
    """
    Decorator for retrying a function if exception occurs
    Source: https://gist.github.com/treo/728327

    tries -- num tries
    exceptions -- exceptions to catch
    delay -- wait between retries
    """
    exceptions_ = exceptions or (Exception, )

    def _retry(fn):
        @wraps(fn)
        def __retry(*args, **kwargs):
            for _ in xrange(tries+1):
                try:
                    return fn(*args, **kwargs)
                except exceptions_, e:
                    log.warning("Exception, retrying...", exc_info=e)
                    time.sleep(delay)
            raise # If no success after tries raise last exception
        return __retry

    return _retry


def write_bars_to_file(bars, filename, tz):
    """Creates CSV file from list of Bar instances"""
    date_format_str = "%Y%m%d %H%M%S"

    rows = [{'DateTime':  bar.datetime.astimezone(tz).strftime(date_format_str),
             'Open':	  bar.open,
             'High':	  bar.high,
             'Low':	      bar.low,
             'Close':	  bar.close,
             'Volume':	  bar.volume,
             } for bar in bars]

    if os.path.exists(filename):
        raise Exception("File already exists!")

    fd = os.popen("gzip > %s" % filename, 'w') if filename.endswith('.gz') else open(filename, 'w')

    with fd:
        csv_writer = csv.DictWriter(fd, ['DateTime', 'Open', 'High', 'Low', 'Close', 'Volume'])
        csv_writer.writeheader()
        csv_writer.writerows(rows)


def get_instruments_from_file(filename):
    """Load index from txt file"""
    instruments = []
    with open(filename, 'r') as f:
        for instrument in f:
            instruments.append(instrument.rstrip())
    return instruments



#
import setuptools

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()


setuptools.setup(name='iqfeed',
                 version='0.4.2',
                 description='IQFeed / DTN Data downloader',
                 long_description=long_description,
                 classifiers=[
                             'Development Status :: 4 - Beta',
                             'License :: OSI Approved :: Apache Software License',
                             'Programming Language :: Python :: 2.7',
                             'Topic :: Office/Business :: Financial :: Investment',

                 ],
                 url='http://github.com/tibkiss/iqfeed',
                 author='Tibor Kiss',
                 author_email='tibor.kiss@gmail.com',
                 license='Apache License, Version 2.0',
                 packages=setuptools.find_packages(),
                 install_requires=['docopt', 'pytz', 'backports.functools_lru_cache'],
                 entry_points={'console_scripts': ['iqfeed = iqfeed.main:main']},
                 zip_safe=False)
